import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

public final class Mapper020 extends Mapper {
    //private byte bios[];
    private byte wram[][];
    private byte disk[][];
    private byte bios[];

    private byte irq_enabled;
    private int irq_counter;
    private int irq_latch;
    private int irq_wait;

    private boolean disk_enabled;
    private int head_position;
    private int write_skip;
    private byte disk_status;
    private byte write_reg;
    private byte current_side;

    private boolean access_flag;
    private byte last_side;
    private int insert_wait;
    
    // SOUND SETTING
    private byte sound_wav[];
    private byte sound_reg[];

    public Mapper020(final NESCore nes) {
        super(nes);
        wram = new byte[4][0x2002];
        disk = nes.DISK;
        bios = nes.BIOS;

        //sound
        sound_wav = new byte[0x40];
        sound_reg = new byte[0x10];
    }

    public final void Reset() {
        // Init ExSound
        //parent_NES->apu->SelectExSound(4);

        // Memory Clear
        for(int i = 0; i < 4; i++) {
            for(int j = 0; j < 0x2000; j++) wram[i][j] = 0x00;
        }

        //NES_6502::Context context;
        //parent_NES->cpu->GetContext(&context);
        //context.mem_page[3] = wram + 0x0000;
        //context.mem_page[4] = wram + 0x2000;
        //context.mem_page[5] = wram + 0x4000;
        //context.mem_page[6] = wram + 0x6000;
        //context.mem_page[7] = bios;
        //parent_NES->cpu->SetContext(&context);
        nes.cpu_map[3] = wram[0];
        nes.cpu_map[4] = wram[1];
        nes.cpu_map[5] = wram[2];
        nes.cpu_map[6] = wram[3];
        nes.cpu_map[7] = bios;
                
        //if(nes.fds_id() == 0xc7525744) { // Reflect World
        //    nes.frame_irq_enabled = 0;
        //}
        
        irq_enabled = 0;
        irq_counter = 0;
        irq_latch = 0;
        irq_wait = 0;
        
        access_flag = false;
        disk_enabled = true;
        head_position = 0;
        write_skip = 0;
        disk_status = 0;
        write_reg = 0;
        current_side = 1;
        last_side = 1;
        insert_wait = 180;
    }

    public void StateLoad(DataInputStream dis) throws IOException {
        irq_enabled   = dis.readByte();
        irq_counter   = dis.readInt();
        irq_latch     = dis.readInt();
        irq_wait      = dis.readInt();
        disk_enabled  = dis.readBoolean();
        head_position = dis.readInt();
        write_skip    = dis.readInt();
        disk_status   = dis.readByte();
        write_reg     = dis.readByte();
        current_side  = dis.readByte();
        access_flag   = dis.readBoolean();
        last_side     = dis.readByte();
        insert_wait   = dis.readInt();
    }

    public void StateSave(DataOutputStream dos) throws IOException {
        dos.writeByte(irq_enabled);
        dos.writeInt(irq_counter);
        dos.writeInt(irq_latch);
        dos.writeInt(irq_wait);
        dos.writeBoolean(disk_enabled);
        dos.writeInt(head_position);
        dos.writeInt(write_skip);
        dos.writeByte(disk_status);
        dos.writeByte(write_reg);
        dos.writeByte(current_side);
        dos.writeBoolean(access_flag);
        dos.writeByte(last_side);
        dos.writeInt(insert_wait);
    }

    public final int MemoryReadLow(final int addr) {
        int val = 0x00;
        
        switch (addr) {
        case 0x4030:
        {
            val = disk_status & 0xFF;
        }
        break;
        case 0x4031:
        {
            if((current_side != 0) && (current_side == last_side)) {
                val = disk[current_side - 1][head_position] & 0xFF;
                if((write_reg & 0x01) != 0) {
                    head_position += (head_position < 64999) ? 1 : 0;
                    irq_wait = 2;
                }
                access_flag = true;
            } else {
                val = 0xff;
            }
        }
        break;
        case 0x4032:
        {
            boolean eject = ((current_side == 0) || (current_side != last_side));
            val = 0x40;
            val |= eject ? 1 : 0;
            val |= eject ? 4 : 0;
            val |= ((!eject) && ((write_reg & 0x01)!=0) && ((write_reg & 0x02)==0)) ? 0 : 2;
            if(last_side != current_side) {
                // wait 2.0 sec for change disk
                if(insert_wait > 120) {
                    last_side = current_side;
                }
            }
        }
        break;
        case 0x4033:
        {
            val = 0x80;
        }
        break;
        default:
        {
            if(addr >= 0x4040 && addr < 0x4080) {
                return sound_wav[addr - 0x4040] & 0xFF;
            } else if (addr >= 0x4090 && addr < 0x40A0) {
                return sound_reg[addr - 0x4090] & 0xFF;
            }

            // XXX
            //val = parent_NES.apu.ExRead(addr);
            //System.out.println("ExRead(" + Integer.toHexString(addr) + ")");
            val = 0;
        }
        break;
        }
        return val;
    }

    public void MemoryWriteLow(int addr, int data) {
        switch (addr) {
        case 0x4020:
        {
            irq_latch = (irq_latch & 0xFF00) | data;
        }
        break;
        case 0x4021:
        {
            irq_latch = (irq_latch & 0x00FF) | (data << 8);
        }
        break;
        case 0x4022:
        {
            irq_counter = irq_latch;
            irq_enabled = (byte)(data & 0x03);
        }
        break;
        case 0x4023:
        {
            disk_enabled = ((data & 0x01) != 0);
        }
        break;
        case 0x4024:
        {
            if((current_side != 0) && (current_side == last_side)) {
                if((disk_enabled) && ((write_reg & 0x04) == 0) && head_position < 65000) {
                    if(write_skip != 0) {
                        write_skip--;
                    } else if(head_position >= 2) {
                        disk[current_side - 1][head_position - 2] = (byte)data;
                        access_flag = true;
                    }
                }
            }
        }
        break;
        case 0x4025:
        {
            if((data & 0x08)!=0) {
                set_mirroring(NESCore.MIRROR_HORIZ);
            } else {
                set_mirroring(NESCore.MIRROR_VERT);
            }
            if((current_side != 0) && (current_side == last_side)) {
                if(((write_reg & 0x40) != 0) && ((data & 0x10) == 0) && ((data & 0x40) == 0)) {
                    head_position = (head_position < 2) ? 0 : head_position - 2;
                }
                if((data & 0x04) == 0) {
                    write_skip = 2;
                }
                if((data & 0x02) != 0) {
                    head_position = 0;
                    irq_wait = 2;
                }
                if((data & 0x80) != 0) {
                    irq_wait = 2;
                }
            }
            write_reg = (byte)data;
        }
        break;
        default:
        {
            if (addr >= 0x4040 && addr <= 0x407F) {
                sound_wav[addr - 0x4040] = (byte)data;
            } else if (0x4080 <= addr && addr <= 0x408F) {
                sound_reg[addr - 0x4080] = (byte)data;
            }

            // XXX
            //parent_NES.apu.ExWrite(addr, data);
            //System.out.println("ExWrite(" + Integer.toHexString(addr) + ")");
        }
        break;
        }
    }

    public final void MemoryWriteSaveRAM(int addr, int data) {
        wram[(addr >> 13) - 3][addr & 0x1FFF] = (byte)data;
        if((addr >> 13) > 3 &&(addr & 0x1FFE) == 0){
            wram[(addr >> 13) - 4][0x2000 + (addr & 0x0001)] = (byte)data;
        }
    }

    public final void MemoryWrite(int addr, int data) {
        if(addr < 0xE000) {
            wram[(addr >> 13) - 3][addr & 0x1FFF] = (byte)data;
            if((addr >> 13) > 3 &&(addr & 0x1FFE) == 0){
                wram[(addr >> 13) - 4][0x2000 + (addr & 0x0001)] = (byte)data;
            }
        }
    }

    public final void HSync(int scanline) {
        disk_status &= 0xfc;
        if(irq_enabled != 0) {
            if(irq_counter < 113) {
                irq_enabled &= 0x01;
                irq_counter = irq_latch;
                disk_status |= 0x01;
                nes.do_irq();
            } else {
                irq_counter -= 113;
            }
        } else if(irq_wait != 0) {
            irq_wait--;
            if((irq_wait == 0) && ((write_reg & 0x80) != 0)) {
                disk_status |= 0x02;
                nes.do_irq();
            }
        }
    }

    public final void VSync() {
        // count MAX 3sec
        insert_wait += (insert_wait < 180) ? 1 : 0;
    }

    public final byte GetDiskSideNum() {
        return (byte)nes.num_DISK_banks;
    }

    public final byte GetDiskSide() {
        return current_side;
    }

    public final void SetDiskSide(byte side) {
        if(side > 0 && side <= nes.num_DISK_banks) {
            current_side = side;
            insert_wait = 0;
        }
    }

    public byte GetDiskData(int pt) {
        return disk[current_side - 1][pt];
    }

    public final void SetDiskData(int pt, byte data) {
        disk[current_side - 1][pt] = data;
    }

    public final byte DiskAccessed() {
        byte retval = (access_flag)?((byte)1):((byte)0);
        access_flag = false;
        return retval;
    }

    public final void SNSS_fixup() {
        //NES_6502::Context context;
        //parent_NES.cpu.GetContext(context);
        //context.mem_page[3] = wram + 0x0000;
        //context.mem_page[4] = wram + 0x2000;
        //context.mem_page[5] = wram + 0x4000;
        //context.mem_page[6] = wram + 0x6000;
        //context.mem_page[7] = bios;
        //parent_NES.cpu.SetContext(context);
        nes.cpu_map[3] = wram[0];
        nes.cpu_map[4] = wram[1];
        nes.cpu_map[5] = wram[2];
        nes.cpu_map[6] = wram[3];
        nes.cpu_map[7] = bios;
    }
}
